package net.pingex.discordbot;

import sx.blah.discord.api.EventSubscriber;
import sx.blah.discord.handle.impl.events.MessageReceivedEvent;
import sx.blah.discord.util.DiscordException;
import sx.blah.discord.util.HTTP429Exception;
import sx.blah.discord.util.MissingPermissionsException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * This class parses and dispatch Discord commands
 * @version 0.1-dev
 * @author Raphael "Pingex" NAAS
 */
public class CommandDispatcher
{
	private Logger logger;

	/**
	 * Contains all available commands, built using `rebuildCommandList()`
	 */
	private HashMap<String, InvokableMethod> commandList;

	/**
	 * Basic Constructor, automatically rebuilds command list from the modules registry
	 */
	public CommandDispatcher()
	{
		logger = Logger.getLogger(this.getClass().getName());
		rebuildCommandList();
	}

	/**
	 * Fired when receiving a message
	 * @param event Event from Discord API
	 */
	@EventSubscriber
	public void onMessageReceivedEvent(MessageReceivedEvent event)
	{
		Matcher m = Pattern.compile("^!(\\w+):(\\w+)(?: (.*))?$").matcher(event.getMessage().getContent()); // TODO: Custom command prefix
		if(!m.matches()) return; // We don't need to go further if it's not even a command

		String module = m.group(1);
		String command = m.group(2);
		String fullCommand = module + ":" + command;

		// Arg splitter
		Matcher am = Pattern.compile("([^\"]\\S*|\".+?\")\\s*").matcher((m.group(3) != null) ? m.group(3) : "");
		ArrayList<String> args = new ArrayList<>();
		while(am.find()) args.add(am.group(1).replace("\"", ""));

		logger.info("Command invoked (" + event.getMessage().getAuthor().getName() + "#" + event.getMessage().getAuthor().getDiscriminator() + "): " + fullCommand);

		try
		{
			if(commandList.containsKey(fullCommand))
			{
				InvokableMethod foundMethod = commandList.get(fullCommand);
				Object[] parsedArray = new Object[foundMethod.getMethod().getParameterCount()];
				parsedArray[0] = event;

				if(foundMethod.getMethod().getParameterCount()-1 != args.size())
				{
					event.getMessage().reply("Invalid arguments. " + getHelp(fullCommand));
					return;
				}

				for(int i=1; i < foundMethod.getMethod().getParameterCount(); i++)
					try
					{
						parsedArray[i] = parse(foundMethod.getMethod().getParameterTypes()[i], args.get(i-1));
					} catch (IllegalArgumentException e)
					{
						event.getMessage().reply("Failed to parse arguments, are they correct ? " + getHelp(fullCommand));
					}

				String answer = (String) commandList.get(fullCommand).invoke(parsedArray);
				if(answer != null) event.getMessage().reply(answer);
			}
			else
				event.getMessage().reply("Unknown command");
		} catch (MissingPermissionsException | HTTP429Exception | DiscordException e)
		{
			logger.warning("Couldn't reply to command (" + e.getClass().getName() + "): " + e.getMessage());
		} catch (IllegalArgumentException | IllegalAccessException | InvocationTargetException e)
		{
			logger.severe("Couldn't call target method (" + e.getClass().getName() + "): " + e.getMessage());
		}
	}

	/**
	 * Rebuilds command list using the module registry.
	 * - It looks for methods with @Command annotation inside class with @Controllable annotation
	 * - Check his prototype (should be `method(String[]):String`)
	 * - Add usable methods in his local registry, using InvokableMethod
	 * @see InvokableMethod
	 * @see ModulesRegistry
	 */
	public void rebuildCommandList()
	{
		logger.info("Rebuilding command list...");

		ArrayList<AbstractModule> registry = ModulesRegistry.getRegistry();
		commandList = new HashMap<>();

		for(AbstractModule i : registry)
			if(i.getClass().isAnnotationPresent(Controllable.class))
				for(Method j : i.getClass().getDeclaredMethods())
					if(j.isAnnotationPresent(Command.class))
					{
						String id = i.getClass().getAnnotation(Controllable.class).name() + ":" + j.getName();
						logger.info("Found " + id);

						if(j.getParameterCount() >= 1 && j.getParameterTypes()[0] == MessageReceivedEvent.class && j.getReturnType() == String.class)
							commandList.put(id, new InvokableMethod(j, i));
						else
							logger.warning("Command [" + id + "]: incorrect function prototype, thus won't be added to the command list.");
					}

		logger.info("... Done");
	}

	/**
	 * Parses a String into a primitive variable. Choice between boolean, byte, short, int, long, float, double, string
	 * @param target Target primitive
	 * @param value Input value
	 * @return `null` if the value is null or the string content is "null", or the parsed value.
	 * @throws IllegalArgumentException If the target primitive/class isn't part of the previous list
	 */
	public static Object parse(Class<?> target, String value) throws IllegalArgumentException
	{
		if(value         == null   || value.equals("null")  ) return null;
		if(Boolean.class == target || Boolean.TYPE == target) return Boolean.parseBoolean(value);
		if(Byte.class    == target || Byte.TYPE    == target) return Byte.parseByte(value);
		if(Short.class   == target || Short.TYPE   == target) return Short.parseShort(value);
		if(Integer.class == target || Integer.TYPE == target) return Integer.parseInt(value);
		if(Long.class    == target || Long.TYPE    == target) return Long.parseLong(value);
		if(Float.class   == target || Float.TYPE   == target) return Float.parseFloat(value);
		if(Double.class  == target || Double.TYPE  == target) return Double.parseDouble(value);
		if(String.class  == target)                           return value;
		throw new IllegalArgumentException("Unknown target type");
	}

	/**
	 * Scans and print the usage of a command
	 * @param command Full command `module:command`
	 * @return Human-readable string
	 */
	private String getHelp(String command)
	{
		if(!commandList.containsKey(command)) return "No help available for this command.";
		InvokableMethod matchingMethod = commandList.get(command);

		StringBuffer toReturn = new StringBuffer("Usage: ");
		toReturn.append("!" + command + " ");
		for(int i=1; i<matchingMethod.getMethod().getParameterCount(); i++)
			toReturn.append(matchingMethod.getMethod().getParameterTypes()[i].getName() + " ");

		return toReturn.toString();
	}
}
